#!/usr/bin/env python
# crate_anon/crateweb/extra/fields.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import datetime
import dateutil.parser
import logging
import os
import pytz
from typing import Any, Dict, Iterable, List, Optional, Tuple

from django import forms
from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.db.models.fields import DateField, DateTimeField, Field
from django.template.defaultfilters import filesizeformat
from django.utils import timezone
from django.utils.translation import ugettext_lazy

log = logging.getLogger(__name__)


# =============================================================================
# ContentTypeRestrictedFileField
# =============================================================================
# https://djangosnippets.org/snippets/2206/
# https://docs.djangoproject.com/en/1.8/ref/files/uploads/

class ContentTypeRestrictedFileField(models.FileField):
    """
    Same as FileField, but you can specify:
        * content_types - list containing allowed content_types.
          Example: ['application/pdf', 'image/jpeg']
        * max_upload_size - a number indicating the maximum file size allowed
          for upload.
            2.5MB - 2621440
            5MB - 5242880
            10MB - 10485760
            20MB - 20971520
            50MB - 5242880
            100MB - 104857600
            250MB - 214958080
            500MB - 429916160
    """
    def __init__(self, *args, **kwargs) -> None:
        self.content_types = kwargs.pop("content_types", None)
        if self.content_types is None:
            self.content_types = []
        self.max_upload_size = kwargs.pop("max_upload_size", None)
        super().__init__(*args, **kwargs)

    def clean(self, *args, **kwargs) -> Any:
        data = super().clean(*args, **kwargs)
        # log.debug("data: {}".format(repr(data)))
        f = data.file
        if not isinstance(f, UploadedFile):  # RNC
            # no new file uploaded; there won't be a content-type to check
            return data
        # log.debug("f: {}".format(repr(f)))
        content_type = f.content_type
        if content_type not in self.content_types:
            raise forms.ValidationError(ugettext_lazy(
                'Filetype not supported.'))
        # noinspection PyProtectedMember
        if self.max_upload_size is not None and f._size > self.max_upload_size:
            # noinspection PyProtectedMember
            raise forms.ValidationError(ugettext_lazy(
                'Please keep filesize under %s. Current filesize %s')
                % (filesizeformat(self.max_upload_size),
                   filesizeformat(f._size)))
        return data


# =============================================================================
# Making FileFields own their files (i.e. delete them afterwards)
# =============================================================================

# http://stackoverflow.com/questions/16041232/django-delete-filefield
# These two auto-delete files from filesystem when they are unneeded:
# ... with a bit of modification to make them generic (RNC)
# Attach them with signals; see e.g. Study model.
def auto_delete_files_on_instance_delete(instance: Any,
                                         fieldnames: Iterable[str]) -> None:
    """Deletes files from filesystem when object is deleted."""
    for fieldname in fieldnames:
        filefield = getattr(instance, fieldname, None)
        if filefield:
            if os.path.isfile(filefield.path):
                os.remove(filefield.path)


def auto_delete_files_on_instance_change(instance: Any,
                                         fieldnames: Iterable[str],
                                         model: models.Model) -> None:
    """Deletes files from filesystem when object is changed."""
    if not instance.pk:
        return  # instance not yet saved in database
    try:
        old_instance = model.objects.get(pk=instance.pk)
    except model.DoesNotExist:
        return  # old version gone from database entirely
    for fieldname in fieldnames:
        old_filefield = getattr(old_instance, fieldname, None)
        if not old_filefield:
            continue
        new_filefield = getattr(instance, fieldname, None)
        if old_filefield != new_filefield:
            if os.path.isfile(old_filefield.path):
                os.remove(old_filefield.path)


# =============================================================================
# Field choice assistance
# =============================================================================

def valid_choice(strvalue: str, choices: Iterable[Tuple[str, str]]) -> bool:
    """
    Checks that value is one of the valid option in choices, where choices
    is a list/tuple of 2-tuples (option, description).

    Note that parameters sent by URLconf are always strings:
        https://docs.djangoproject.com/en/1.8/topics/http/urls/
    but Python is happy with a string-to-integer-PK lookup, e.g.
        Study.objects.get(pk=1)
        Study.objects.get(pk="1")  # also works

    Choices can be non-string, though, so we compare against a string version
    of the choice.
    """
    return strvalue in [str(x[0]) for x in choices]


def choice_explanation(value: str, choices: Iterable[Tuple[str, str]]) -> str:
    """
    Returns the explanation associated with a Django choice tuple-list.
    """
    for k, v in choices:
        if k == value:
            return v
    return ''


# =============================================================================
# IsoDateTimeTzField
# =============================================================================

# -----------------------------------------------------------------------------
# Conversions: Python
# -----------------------------------------------------------------------------

def iso_string_to_python_datetime(
        isostring: str) -> Optional[datetime.datetime]:
    """Takes an ISO-8601 string and returns a datetime."""
    if not isostring:
        return None  # if you parse() an empty string, you get today's date
    return dateutil.parser.parse(isostring)


def python_utc_datetime_to_sqlite_strftime_string(
        value: datetime.datetime) -> str:
    millisec_str = str(round(value.microsecond / 1000)).zfill(3)
    return value.strftime("%Y-%m-%d %H:%M:%S") + "." + millisec_str


def python_localized_datetime_to_human_iso(value: datetime.datetime) -> str:
    s = value.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
    return s[:29] + ":" + s[29:]


# -----------------------------------------------------------------------------
# Field
# -----------------------------------------------------------------------------

class IsoDateTimeTzField(models.CharField):
    """
    Microsecond resolution; timezone-aware.
    Example sent TO the database:
        2015-11-11T22:21:37.000000+05:00

        YYYY-MM-DD HH:MM:SS uuuuuu TZ:TZ
        1234567890123456789012345678901234567890 ... so 32 chars

    Will also accept FROM the database any ISO-8601 format accepted by
    dateutil.parser.parse()

    - Python side: datetime.datetime

    - The database-specific bits are a tricky.

        In SQLite, the DATETIME() and STRFTIME() functions can be used to
        convert ISO strings to UTC, so that's a good basis for comparison.
        However, the actual comparisons, as always in SQLite, are string-based.

        DATETIME(...) ends up like this:
            2015-11-11 22:21:37

        STRFTIME('%Y-%m-%d %H:%M:%f', ...) looks like this, allowing
        millisecond precision:
            2015-11-11 22:21:37.000
            2015-11-11 22:21:37.123

        The Django automatic RHS converter for datetime values looks like this:
            2015-11-11 22:21:37
            2015-11-11 22:21:37.123456

            ... http://stackoverflow.com/questions/14368290/django-datetimefield-and-timezone-now  # noqa

        ... so a direct comparison with DATETIME works only if the microsecond
        component is zero (or you coerce it to zero via get_db_prep_value), and
        a direct comparison to the STRFTIME expression fails.

        Coercing the fractional seconds to zero via get_db_prep_value() would
        make sub-second comparisons meaningless in the database. So we should
        avoid that, and operate as close to the database resolution as
        possible.

        That means that we have to alter the RHS in the same way we altered the
        LHS, and that means we can't use a transform...
        ... oh, hang on, I'm just underestimating get_db_prep_value(); we can
        coerce to exactly the thing yielded by SQLite's STRFTIME, by converting
        to a string!

        Next step: the data conversion specified for the RHS of a transform is
        governed by its output_field. So when we're coercing to a date, we
        specify a DateField here, and all is happy. But for a datetime, we
        should NOT specify DateTimeField, because that will bypass the equality
        lookup from our custom field, and also bypass the RHS conversion (via
        get_db_prep_value() etc.).

        Specifically, for a DateTimeField, conversion to a database value is
        done by
            django/db/models/fields/__init__.py : DateTimeField.get_prep_value
            django/db/backends/{backend}/operations.py
                : DatabaseOperations.value_to_db_datetime
                - for sqlite3: calls str(value) [via six.texttype]
                  and for a datetime, that gives
                    2015-11-17 00:00:00  # if microsecond == 0
                    2015-11-17 00:00:00.001000  # if microsecond != 0
            So for a proper conversion, we need to convert SQLite stuff to
            that. The MySQL backend does the same thing, but MySQL has a
            concept of a datetime anyway.

    - For MySQL, see also:

        https://docs.djangoproject.com/en/1.8/ref/databases/#fractional-seconds-support-for-time-and-datetime-fields  # noqa

    """
    # https://docs.djangoproject.com/en/1.8/ref/models/fields/#field-api-reference  # noqa

    description = "ISO-8601 date/time field with timezone, stored as text"

    def __init__(self, *args, **kwargs) -> None:
        """Declare that we're a VARCHAR(32) on the database side."""
        # https://docs.djangoproject.com/en/1.8/howto/custom-model-fields/
        kwargs['max_length'] = 32
        super().__init__(*args, **kwargs)

    def deconstruct(self) -> Tuple[str, str, List[Any], Dict[str, Any]]:
        """Takes an instance and calculates the arguments to pass to __init__
        to reconstruct it."""
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']
        return name, path, args, kwargs

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def from_db_value(self, value, expression, connection, context):
        """
        Convert database value to Python value.
        Called when data is loaded from the database.
        """
        # log.debug("from_db_value: {}, {}".format(value, type(value)))
        if value is None:
            return value
        if value == '':
            return None
        return iso_string_to_python_datetime(value)

    def to_python(self, value: Optional[str]) -> Optional[Any]:
        """
        Called during deserialization and during form clean() calls.
        Must deal with an instance of the correct type; a string; or None
        (if the field allows null=True).
        Should raise ValidationError if problems.
        """
        # https://docs.djangoproject.com/en/1.8/howto/custom-model-fields/
        # log.debug("to_python: {}, {}".format(value, type(value)))
        if isinstance(value, datetime.datetime):
            return value
        if value is None:
            return value
        if value == '':
            return None
        return iso_string_to_python_datetime(value)

    def get_prep_value(self, value):
        """
        Convert Python value to database value for QUERYING.
        We query with UTC, so this function converts datetime values to UTC.

        Calls to this function are followed by calls to get_db_prep_value(),
        which is for backend-specific conversions.
        """
        log.debug("get_prep_value: {}, {}".format(value, type(value)))
        if not value:
            return ''
            # For underlying (database) string types, e.g. VARCHAR, this
            # function must always return a string type.
            # https://docs.djangoproject.com/en/1.8/howto/custom-model-fields/
        # Convert to UTC
        return value.astimezone(timezone.utc)

    def get_db_prep_value(self, value, connection, prepared=False):
        """
        Further conversion of Python value to database value for QUERYING.
        This follows get_prep_value(), and is for backend-specific stuff.
        See notes above.
        """
        log.debug("get_db_prep_value: {}, {}".format(value, type(value)))
        value = super().get_db_prep_value(value, connection, prepared)
        if value is None:
            return value
        # log.debug("connection.settings_dict['ENGINE']: {}".format(
        #              connection.settings_dict['ENGINE']))
        if connection.settings_dict['ENGINE'] == 'django.db.backends.sqlite3':
            return python_utc_datetime_to_sqlite_strftime_string(value)
        return value

    def get_db_prep_save(self, value, connection, prepared=False):
        """
        Convert Python value to database value for SAVING.
        We save with full timezone information.
        """
        log.debug("get_db_prep_save: {}, {}".format(value, type(value)))
        if not value:
            return ''
            # For underlying (database) string types, e.g. VARCHAR, this
            # function must always return a string type.
            # https://docs.djangoproject.com/en/1.8/howto/custom-model-fields/
        return python_localized_datetime_to_human_iso(value)


# -----------------------------------------------------------------------------
# Conversions: MySQL
# -----------------------------------------------------------------------------

def iso_string_to_sql_utcdatetime_mysql(x: str) -> str:
    return (
        "CONVERT_TZ(STR_TO_DATE(LEFT({x}, 26),"
        "                       '%Y-%m-%dT%H:%i:%s.%f'),"
        "           RIGHT({x}, 6),"  # from timezone
        "           '+00:00')"  # to timezone
    ).format(x=x)
    # In MySQL:
    # 1. STR_TO_DATE(), with the leftmost 23 characters,
    #    giving microsecond precision, but not correct for timezone
    # 2. CONVERT_TZ(), converting from the timezone info in the rightmost 6
    #    characters to UTC (though losing fractional seconds)


def iso_string_to_sql_utcdate_mysql(x: str) -> str:
    return (
        "DATE(CONVERT_TZ(STR_TO_DATE(LEFT({x}, 26),"
        "                            '%Y-%m-%dT%H:%i:%s.%f'),"
        "                RIGHT({x}, 6),"
        "                '+00:00')"
    ).format(x=x)


def iso_string_to_sql_date_mysql(x: str) -> str:
    return "STR_TO_DATE(LEFT({x}, 10), '%Y-%m-%d')".format(x=x)


# -----------------------------------------------------------------------------
# Conversions: SQLite
# -----------------------------------------------------------------------------

def iso_string_to_sql_utcdatetime_sqlite(x: str) -> str:
    """
    Output like:
        2015-11-14 18:52:47.000
        2015-11-14 18:52:47.247
    Don't use DATETIME; using STRFTIME allows millsecond precision.
    """
    return "STRFTIME('%Y-%m-%d %H:%M:%f', {x})".format(x=x)
    # This doesn't mind the 'T' in the middle, rounds to millisecond precision,
    # and corrects for any timezone at the end without having to tell it to
    # explicitly.
    # Try:
    # SELECT strftime('%Y-%m-%d %H:%M:%f', '2015-11-15T18:29:02.34544+05:00');
    #
    # http://www.sqlite.org/lang_datefunc.html


def iso_string_to_sql_utcdatetime_pythonformat_sqlite(x: str) -> str:
    """
    Output like
        2015-11-14 18:52:47
        2015-11-14 18:52:47.247000
    ... i.e. gets rid of trailing '.000' for zero milliseconds, appends
        trailing '000' for everything else,
    ... thus matching the output of Python's str(x) where x is a datetime.
    ... thus matching the RHS of a Django default datetime comparison.
    """
    return """
        CASE SUBSTR(STRFTIME('%f', {x}), 4, 3)
        WHEN '000' THEN STRFTIME('%Y-%m-%d %H:%M:%S', {x})
        ELSE STRFTIME('%Y-%m-%d %H:%M:%f', {x}) || '000'
        END
    """.format(x=x)


def iso_string_to_sql_utcdate_sqlite(x: str) -> str:
    return "DATE({x})".format(x=x)


def iso_string_to_sql_date_sqlite(x: str) -> str:
    return "DATE(SUBSTR({x}, 1, 10))".format(x=x)


# -----------------------------------------------------------------------------
# Lookups
# -----------------------------------------------------------------------------

def isodt_lookup_mysql(lookup, compiler, connection,
                       operator) -> Tuple[str, Any]:
    lhs, lhs_params = compiler.compile(lookup.lhs)
    rhs, rhs_params = lookup.process_rhs(compiler, connection)
    params = lhs_params + rhs_params
    return '{lhs} {op} {rhs}'.format(
        lhs=iso_string_to_sql_utcdatetime_mysql(lhs),
        op=operator,
        rhs=rhs,
    ), params


def isodt_lookup_sqlite(lookup, compiler, connection,
                        operator) -> Tuple[str, Any]:
    lhs, lhs_params = compiler.compile(lookup.lhs)
    rhs, rhs_params = lookup.process_rhs(compiler, connection)
    params = lhs_params + rhs_params
    return '{lhs} {op} {rhs}'.format(
        lhs=iso_string_to_sql_utcdatetime_sqlite(lhs),
        op=operator,
        rhs=rhs,
    ), params
    # ... RHS conversion using STRFTIME() not necessary because we do the
    # appropriate thing in get_db_prep_value().


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoDateTimeLessThan(models.Lookup):
    lookup_name = 'lt'

    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_mysql(self, compiler, connection, "<")

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_sqlite(self, compiler, connection, "<")


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoDateTimeLessThanEqual(models.Lookup):
    lookup_name = 'lte'

    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_mysql(self, compiler, connection, "<=")

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_sqlite(self, compiler, connection, "<=")


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoDateTimeExact(models.Lookup):
    lookup_name = 'exact'

    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_mysql(self, compiler, connection, "=")

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_sqlite(self, compiler, connection, "=")


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoDateTimeGreaterThan(models.Lookup):
    lookup_name = 'gt'

    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_mysql(self, compiler, connection, ">")

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_sqlite(self, compiler, connection, ">")


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoDateTimeGreaterThanEqual(models.Lookup):
    lookup_name = 'gte'

    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_mysql(self, compiler, connection, ">=")

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        return isodt_lookup_sqlite(self, compiler, connection, ">=")


# -----------------------------------------------------------------------------
# Transforms
# -----------------------------------------------------------------------------

# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoStringToUtcDateTime(models.Transform):
    """
    SQL expression: converts ISO-8601 field into UTC DATETIME.
    """
    lookup_name = 'utc'

    # NOTE THAT SETTING output_field MEANS YOU HAVE TO MAKE THE OUTPUT
    # MATCH THE FORMAT EXPECTED FOR A DateTimeField. The class's own
    # get_db_prep_value(), etc., will NOT be called.
    @property
    def output_field(self) -> Field:
        return DateTimeField()

    # noinspection PyUnusedLocal
    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        log.debug("IsoStringToUtcDateTime.as_mysql")
        lhs, params = compiler.compile(self.lhs)
        return iso_string_to_sql_utcdatetime_mysql(lhs), params

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        log.debug("IsoStringToUtcDateTime.as_sqlite")
        lhs, params = compiler.compile(self.lhs)
        return iso_string_to_sql_utcdatetime_pythonformat_sqlite(lhs), params


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoStringToUtcDate(models.Transform):
    """
    SQL expression: converts ISO-8601 field into DATE, using the UTC date.
    """
    lookup_name = 'utcdate'

    @property
    def output_field(self) -> Field:
        return DateField()

    # noinspection PyUnusedLocal
    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        log.debug("IsoStringToUtcDate.as_mysql")
        lhs, params = compiler.compile(self.lhs)
        return iso_string_to_sql_utcdate_mysql(lhs), params

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        log.debug("IsoStringToUtcDate.as_sqlite")
        lhs, params = compiler.compile(self.lhs)
        return iso_string_to_sql_utcdate_sqlite(lhs), params


# noinspection PyAbstractClass
@IsoDateTimeTzField.register_lookup
class IsoStringToSourceDate(models.Transform):
    """
    SQL expression: converts ISO-8601 field into DATE, using the date part of
    the local ISO time (not the UTC date).
    """
    lookup_name = 'sourcedate'

    @property
    def output_field(self) -> Field:
        return DateField()

    # noinspection PyUnusedLocal
    def as_mysql(self, compiler, connection) -> Tuple[str, Any]:
        log.debug("IsoStringToSourceDate.as_mysql")
        lhs, params = compiler.compile(self.lhs)
        return iso_string_to_sql_date_mysql(lhs), params

    def as_sqlite(self, compiler, connection) -> Tuple[str, Any]:
        log.debug("IsoStringToSourceDate.as_sqlite")
        lhs, params = compiler.compile(self.lhs)
        return iso_string_to_sql_date_sqlite(lhs), params


# -----------------------------------------------------------------------------
# Other
# -----------------------------------------------------------------------------

def get_now_utc() -> datetime.datetime:
    """
    Get the time now in the UTC timezone.
    """
    return datetime.datetime.now(pytz.utc)
    # ... or, in Django-speak, timezone.now().astimezone(timezone.utc)



#!/usr/bin/env python
# crate_anon/anonymise/scrub.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Scrubber classes for CRATE anonymiser.
"""

from collections import OrderedDict
import datetime
import logging
from typing import Any, Dict, Iterable, Generator, List, Optional, Union

from cardinal_pythonlib.rnc_datetime import (
    coerce_to_date,
)
from cardinal_pythonlib.rnc_db import (
    is_sqltype_date,
    is_sqltype_text_over_one_char,
)

from crate_anon.common.hash import GenericHasher
from crate_anon.anonymise.constants import SCRUBMETHOD
from crate_anon.anonymise.anonregex import (
    get_anon_fragments_from_string,
    get_code_regex_elements,
    get_date_regex_elements,
    get_number_of_length_n_regex_elements,
    get_phrase_regex_elements,
    get_regex_from_elements,
    get_regex_string_from_elements,
    get_string_regex_elements,
    get_uk_postcode_regex_elements,
)
from crate_anon.common.stringfunc import (
    get_digit_string_from_vaguely_numeric_string,
    reduce_to_alphanumeric,
)

log = logging.getLogger(__name__)


# =============================================================================
# Generic scrubber base class
# =============================================================================

class ScrubberBase(object):
    """Scrubber base class."""

    def __init__(self, hasher: GenericHasher) -> None:
        """
        :param hasher: object implementing GenericHasher interface
        """
        self.hasher = hasher

    def scrub(self, text: str) -> str:
        """Returns a scrubbed version of the text."""
        raise NotImplementedError()

    def get_hash(self) -> str:
        """Returns a hash of the scrubber itself."""
        raise NotImplementedError()


# =============================================================================
# WordList -- this serves a dual function as a whitelist (is a word in the
# list?) and a blacklist (scrub text using the wordlist).
# =============================================================================

def lower_case_words_from_file(fileobj: Iterable[str]) -> Generator[str, None,
                                                                    None]:
    for line in fileobj:
        for word in line.split():
            yield word.lower()


class WordList(ScrubberBase):
    def __init__(self,
                 filenames: Iterable[str] = None,
                 words: Iterable[str] = None,
                 replacement_text: str = '[---]',
                 hasher: GenericHasher = None,
                 suffixes: List[str] = None,
                 at_word_boundaries_only: bool = True,
                 max_errors: int = 0) -> None:
        filenames = filenames or []
        words = words or []

        super().__init__(hasher)
        self.replacement_text = replacement_text
        self.suffixes = suffixes
        self.at_word_boundaries_only = at_word_boundaries_only
        self.max_errors = max_errors
        self._regex = None
        self._cached_hash = None
        self._regex_built = False

        self.words = set()
        # Sets are faster than lists for "is x in s" operations:
        # http://stackoverflow.com/questions/2831212/python-sets-vs-lists
        # noinspection PyTypeChecker
        for f in filenames:
            self.add_file(f, clear_cache=False)
        # noinspection PyTypeChecker
        for w in words:
            self.add_word(w, clear_cache=False)

    def clear_cache(self) -> None:
        """Clear cached information."""
        self._regex = None
        self._cached_hash = None

    def add_word(self, word: str, clear_cache: bool = True) -> None:
        if not word:
            return
        self.words.add(word.lower())
        if clear_cache:
            self.clear_cache()

    def add_file(self, filename: str, clear_cache: bool = True) -> None:
        with open(filename) as f:
            wordgen = lower_case_words_from_file(f)
            for w in wordgen:
                self.words.add(w)
        if clear_cache:
            self.clear_cache()

    def contains(self, word: str) -> bool:
        return word.lower() in self.words

    def get_hash(self) -> str:
        # A set is unordered.
        # We want the hash to be the same if we have the same words, even if
        # they were entered in a different order, so we need to sort:
        if not self._cached_hash:
            self._cached_hash = self.hasher.hash(sorted(self.words))
        return self._cached_hash

    def scrub(self, text: str) -> str:
        if not self._regex_built:
            self.build_regex()
        if not self._regex:
            return text
        return self._regex.sub(self.replacement_text, text)

    def build_regex(self) -> None:
        elements = []
        for w in self.words:
            elements.extend(get_string_regex_elements(
                w,
                suffixes=self.suffixes,
                at_word_boundaries_only=self.at_word_boundaries_only,
                max_errors=self.max_errors
            ))
        self._regex = get_regex_from_elements(elements)
        self._regex_built = True


# =============================================================================
# NonspecificScrubber
# Scrubs a bunch of things that are independent of any patient-specific data,
# such as removing all UK postcodes, or numbers of a certain length.
# =============================================================================

class NonspecificScrubber(ScrubberBase):
    def __init__(self,
                 replacement_text: str,
                 hasher: GenericHasher,
                 anonymise_codes_at_word_boundaries_only: bool = True,
                 anonymise_numbers_at_word_boundaries_only: bool = True,
                 blacklist: WordList = None,
                 scrub_all_numbers_of_n_digits: List[int] = None,
                 scrub_all_uk_postcodes: bool = False) -> None:
        """
        scrub_all_numbers_of_n_digits: list of values of n
        """
        scrub_all_numbers_of_n_digits = scrub_all_numbers_of_n_digits or []

        super().__init__(hasher)
        self.replacement_text = replacement_text
        self.anonymise_codes_at_word_boundaries_only = (
            anonymise_codes_at_word_boundaries_only)
        self.anonymise_numbers_at_word_boundaries_only = (
            anonymise_numbers_at_word_boundaries_only)
        self.blacklist = blacklist
        self.scrub_all_numbers_of_n_digits = scrub_all_numbers_of_n_digits
        self.scrub_all_uk_postcodes = scrub_all_uk_postcodes

        self._cached_hash = None
        self._regex = None
        self._regex_built = False
        self.build_regex()

    def get_hash(self) -> str:
        if not self._cached_hash:
            self._cached_hash = self.hasher.hash([
                # signature, used for hashing:
                self.anonymise_codes_at_word_boundaries_only,
                self.anonymise_numbers_at_word_boundaries_only,
                self.blacklist.get_hash() if self.blacklist else None,
                self.scrub_all_numbers_of_n_digits,
                self.scrub_all_uk_postcodes,
            ])
        return self._cached_hash

    def scrub(self, text: str) -> str:
        if not self._regex_built:
            self.build_regex()
        if self.blacklist:
            text = self.blacklist.scrub(text)
        if not self._regex:  # possible; may be blank
            return text
        return self._regex.sub(self.replacement_text, text)

    def build_regex(self) -> None:
        elements = []
        if self.scrub_all_uk_postcodes:
            elements.extend(
                get_uk_postcode_regex_elements(
                    at_word_boundaries_only=
                    self.anonymise_codes_at_word_boundaries_only))
        # noinspection PyTypeChecker
        for n in self.scrub_all_numbers_of_n_digits:
            elements.extend(get_number_of_length_n_regex_elements(
                n,
                at_word_boundaries_only=(
                    self.anonymise_numbers_at_word_boundaries_only)
            ))
        self._regex = get_regex_from_elements(elements)
        self._regex_built = True


# =============================================================================
# PersonalizedScrubber
# =============================================================================

class PersonalizedScrubber(ScrubberBase):
    """Accepts patient-specific (patient and third-party) information, and
    uses that to scrub text."""
    def __init__(self,
                 replacement_text_patient: str,
                 replacement_text_third_party: str,
                 hasher: GenericHasher,
                 anonymise_codes_at_word_boundaries_only: bool = True,
                 anonymise_dates_at_word_boundaries_only: bool = True,
                 anonymise_numbers_at_word_boundaries_only: bool = True,
                 anonymise_numbers_at_numeric_boundaries_only: bool = True,
                 anonymise_strings_at_word_boundaries_only: bool = True,
                 min_string_length_for_errors: int = 4,
                 min_string_length_to_scrub_with: int = 3,
                 scrub_string_suffixes: List[str] = None,
                 string_max_regex_errors: int = 0,
                 whitelist: WordList = None,
                 nonspecific_scrubber: NonspecificScrubber = None,
                 debug: bool = False) -> None:
        scrub_string_suffixes = scrub_string_suffixes or []

        super().__init__(hasher)
        self.replacement_text_patient = replacement_text_patient
        self.replacement_text_third_party = replacement_text_third_party
        self.anonymise_codes_at_word_boundaries_only = (
            anonymise_codes_at_word_boundaries_only)
        self.anonymise_dates_at_word_boundaries_only = (
            anonymise_dates_at_word_boundaries_only)
        self.anonymise_numbers_at_word_boundaries_only = (
            anonymise_numbers_at_word_boundaries_only)
        self.anonymise_numbers_at_numeric_boundaries_only = (
            anonymise_numbers_at_numeric_boundaries_only)
        self.anonymise_strings_at_word_boundaries_only = (
            anonymise_strings_at_word_boundaries_only)
        self.min_string_length_for_errors = min_string_length_for_errors
        self.min_string_length_to_scrub_with = min_string_length_to_scrub_with
        self.scrub_string_suffixes = scrub_string_suffixes
        self.string_max_regex_errors = string_max_regex_errors
        self.whitelist = whitelist
        self.nonspecific_scrubber = nonspecific_scrubber
        self.debug = debug

        # Regex information
        self.re_patient = None  # re: regular expression
        self.re_tp = None
        self.regexes_built = False
        self.re_patient_elements = []
        self.re_tp_elements = []
        # ... both changed from set to list to reflect referee's point re
        #     potential importance of scrubber order
        self.elements_tuplelist = []  # of tuples: (patient?, type, value)
        # ... used for get_raw_info(); since we've made the order important,
        #     we should detect changes in order here as well
        self.clear_cache()

    def clear_cache(self) -> None:
        self.regexes_built = False

    @staticmethod
    def get_scrub_method(datatype_long: str,
                         scrub_method: Optional[SCRUBMETHOD]) -> SCRUBMETHOD:
        """
        Return the default scrub method for a given SQL datatype,
        unless overridden.
        """
        if scrub_method is not None:
            return scrub_method
        elif is_sqltype_date(datatype_long):
            return SCRUBMETHOD.DATE
        elif is_sqltype_text_over_one_char(datatype_long):
            return SCRUBMETHOD.WORDS
        else:
            return SCRUBMETHOD.NUMERIC

    def add_value(self,
                  value: Any,
                  scrub_method: SCRUBMETHOD,
                  patient: bool = True,
                  clear_cache: bool = True) -> None:
        """
        Add a specific value via a specific scrub_method.

        The patient flag controls whether it's treated as a patient value or
        a third-party value.
        """
        if value is None:
            return
        new_tuple = (patient, scrub_method, repr(value))
        if new_tuple not in self.elements_tuplelist:
            self.elements_tuplelist.append(new_tuple)
        # Note: object reference
        r = self.re_patient_elements if patient else self.re_tp_elements

        if scrub_method is SCRUBMETHOD.DATE:
            elements = self.get_elements_date(value)
        elif scrub_method is SCRUBMETHOD.WORDS:
            elements = self.get_elements_words(value)
        elif scrub_method is SCRUBMETHOD.PHRASE:
            elements = self.get_elements_phrase(value)
        elif scrub_method is SCRUBMETHOD.NUMERIC:
            elements = self.get_elements_numeric(value)
        elif scrub_method is SCRUBMETHOD.CODE:
            elements = self.get_elements_code(value)
        else:
            raise ValueError("Bug: unknown scrub_method to add_value: "
                             "{}".format(scrub_method))
        r.extend(elements)
        if clear_cache:
            self.clear_cache()

    def get_elements_date(self,
                          value: Union[datetime.datetime,
                                       datetime.date]) -> Optional[List[str]]:
        # Source is a date.
        try:
            value = coerce_to_date(value)
        except Exception as e:
            log.warning(
                "Invalid date received to PersonalizedScrubber."
                "get_elements_date(): value={}, exception={}".format(
                    value, e))
            return
        return get_date_regex_elements(
            value,
            at_word_boundaries_only=(
                self.anonymise_dates_at_word_boundaries_only)
        )

    def get_elements_words(self, value: str) -> List[str]:
        # Source is a string containing textual words.
        elements = []
        for s in get_anon_fragments_from_string(str(value)):
            l = len(s)
            if l < self.min_string_length_to_scrub_with:
                # With numbers: if you use the length limit, you may see
                # numeric parts of addresses, e.g. 4 Drury Lane as
                # 4 [___] [___]. However, if you exempt numbers then you
                # mess up a whole bunch of quantitative information, such
                # as "the last 4-5 years" getting wiped to "the last
                # [___]-5 years". So let's apply the length limit
                # consistently.
                continue
            if self.whitelist and self.whitelist.contains(s):
                continue
            if l >= self.min_string_length_for_errors:
                max_errors = self.string_max_regex_errors
            else:
                max_errors = 0
            elements.extend(get_string_regex_elements(
                s,
                self.scrub_string_suffixes,
                max_errors=max_errors,
                at_word_boundaries_only=(
                    self.anonymise_strings_at_word_boundaries_only)
            ))
        return elements

    def get_elements_phrase(self, value: Any) -> List[str]:
        # value = unicode(value)  # Python 2
        value = str(value)
        if not value:
            return []
        l = len(value)
        if l < self.min_string_length_to_scrub_with:
            return []
        if self.whitelist and self.whitelist.contains(value):
            return []
        if l >= self.min_string_length_for_errors:
            max_errors = self.string_max_regex_errors
        else:
            max_errors = 0
        return get_phrase_regex_elements(
            value,
            max_errors=max_errors,
            at_word_boundaries_only=(
                self.anonymise_strings_at_word_boundaries_only)
        )

    def get_elements_numeric(self, value: Any) -> List[str]:
        # Source is a text field containing a number, or an actual number.
        # Remove everything but the digits
        # Particular examples: phone numbers, e.g. "(01223) 123456".
        return get_code_regex_elements(
            get_digit_string_from_vaguely_numeric_string(str(value)),
            at_word_boundaries_only=(
                self.anonymise_numbers_at_word_boundaries_only),
            at_numeric_boundaries_only=(
                self.anonymise_numbers_at_numeric_boundaries_only)
        )

    def get_elements_code(self, value: Any) -> List[str]:
        # Source is a text field containing an alphanumeric code.
        # Remove whitespace.
        # Particular examples: postcodes, e.g. "PE12 3AB".
        return get_code_regex_elements(
            reduce_to_alphanumeric(str(value)),
            at_word_boundaries_only=(
                self.anonymise_codes_at_word_boundaries_only)
        )

    def get_patient_regex_string(self) -> str:
        """Return the string version of the patient regex, sorted."""
        return get_regex_string_from_elements(self.re_patient_elements)

    def get_tp_regex_string(self) -> str:
        """Return the string version of the third-party regex, sorted."""
        return get_regex_string_from_elements(self.re_tp_elements)

    def build_regexes(self) -> None:
        self.re_patient = get_regex_from_elements(self.re_patient_elements)
        self.re_tp = get_regex_from_elements(self.re_tp_elements)
        self.regexes_built = True
        # Note that the regexes themselves may be None even if they have
        # been built.
        if self.debug:
            log.debug("Patient scrubber: {}".format(
                self.get_patient_regex_string()))
            log.debug("Third party scrubber: {}".format(
                self.get_tp_regex_string()))

    def scrub(self, text: str) -> Optional[str]:
        """Scrub some text and return the scrubbed result."""
        if text is None:
            return None
        if not self.regexes_built:
            self.build_regexes()

        if self.nonspecific_scrubber:
            text = self.nonspecific_scrubber.scrub(text)
        if self.re_patient:
            text = self.re_patient.sub(self.replacement_text_patient, text)
        if self.re_tp:
            text = self.re_tp.sub(self.replacement_text_third_party, text)
        return text

    def get_hash(self) -> str:
        return self.hasher.hash(self.get_raw_info())

    def get_raw_info(self) -> Dict[str, Any]:
        # This is both a summary for debugging and the basis for our
        # change-detection hash (and for the latter reason we need order etc.
        # to be consistent). For anything we put in here, changes will cause
        # data to be re-scrubbed.
        # We use a list of tuples to make the OrderedDict:
        d = (
            ('anonymise_codes_at_word_boundaries_only',
             self.anonymise_codes_at_word_boundaries_only),
            ('anonymise_dates_at_word_boundaries_only',
             self.anonymise_dates_at_word_boundaries_only),
            ('anonymise_numbers_at_word_boundaries_only',
             self.anonymise_numbers_at_word_boundaries_only),
            ('anonymise_strings_at_word_boundaries_only',
             self.anonymise_strings_at_word_boundaries_only),
            ('min_string_length_for_errors',
             self.min_string_length_for_errors),
            ('min_string_length_to_scrub_with',
             self.min_string_length_to_scrub_with),
            ('scrub_string_suffixes', sorted(self.scrub_string_suffixes)),
            ('string_max_regex_errors', self.string_max_regex_errors),
            ('whitelist_hash',
             self.whitelist.get_hash() if self.whitelist else None),
            ('nonspecific_scrubber_hash',
             self.nonspecific_scrubber.get_hash() if self.nonspecific_scrubber
             else None),
            ('elements', self.elements_tuplelist),
        )
        return OrderedDict(d)
